"""Executor module, implementing facilities for executing quantum programs using Classiq platform."""

from typing import Tuple, Union

from typing_extensions import TypeAlias

from classiq.interface.backend.backend_preferences import BackendPreferencesTypes
from classiq.interface.executor.estimation import OperatorsEstimation
from classiq.interface.executor.execution_preferences import ExecutionPreferences
from classiq.interface.executor.quantum_code import QuantumCode
from classiq.interface.executor.quantum_instruction_set import QuantumInstructionSet
from classiq.interface.executor.result import ExecutionDetails
from classiq.interface.generator.quantum_program import QuantumProgram

from classiq._internals import async_utils
from classiq._internals.api_wrapper import ApiWrapper
from classiq.execution.jobs import ExecutionJob
from classiq.synthesis import SerializedQuantumProgram

BatchExecutionResult: TypeAlias = Union[ExecutionDetails, BaseException]
ProgramAndResult: TypeAlias = Tuple[QuantumCode, BatchExecutionResult]
BackendPreferencesAndResult: TypeAlias = Tuple[
    BackendPreferencesTypes, int, BatchExecutionResult
]


def _parse_serialized_qprog(
    quantum_program: SerializedQuantumProgram,
) -> QuantumProgram:
    return QuantumProgram.parse_raw(quantum_program)


async def execute_async(quantum_program: SerializedQuantumProgram) -> ExecutionJob:
    circuit = _parse_serialized_qprog(quantum_program)
    result = await ApiWrapper.call_execute_generated_circuit(circuit)
    return ExecutionJob(details=result)


def execute(quantum_program: SerializedQuantumProgram) -> ExecutionJob:
    """
    Execute a quantum program. The preferences for execution are set on the quantum program using the method `set_execution_preferences`.

    Args:
        quantum_program: The quantum program to execute. This is the result of the synthesize method.

    Returns:
        ExecutionJob: The result of the execution.

    For examples please see [Execution Documentation](https://docs.classiq.io/latest/reference-manual/executor/)
    """
    return async_utils.run(execute_async(quantum_program))


def set_quantum_program_execution_preferences(
    quantum_program: SerializedQuantumProgram,
    preferences: ExecutionPreferences,
) -> SerializedQuantumProgram:
    circuit = _parse_serialized_qprog(quantum_program)
    circuit.model.execution_preferences = preferences
    return SerializedQuantumProgram(circuit.json())


__all__ = [
    "QuantumCode",
    "QuantumInstructionSet",
    "OperatorsEstimation",
    "set_quantum_program_execution_preferences",
]
