from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, List, Sequence, TypeVar, cast

from classiq.interface.debug_info.debug_info import FunctionDebugInfo
from classiq.interface.ide.visual_model import OperationLevel
from classiq.interface.model.classical_parameter_declaration import (
    ClassicalParameterDeclaration,
)
from classiq.interface.model.handle_binding import HandleBinding
from classiq.interface.model.native_function_definition import NativeFunctionDefinition
from classiq.interface.model.port_declaration import PortDeclaration
from classiq.interface.model.quantum_function_call import ArgValue, QuantumFunctionCall
from classiq.interface.model.quantum_function_declaration import (
    NamedParamsQuantumFunctionDeclaration,
    PositionalArg,
)
from classiq.interface.model.quantum_statement import QuantumStatement
from classiq.interface.model.variable_declaration_statement import (
    VariableDeclarationStatement,
)

from classiq.model_expansions.capturing.propagated_var_stack import (
    PropagatedVarStack,
    validate_args_are_not_propagated,
)
from classiq.model_expansions.closure import FunctionClosure
from classiq.model_expansions.evaluators.argument_types import (
    add_information_from_output_arguments,
)
from classiq.model_expansions.evaluators.parameter_types import (
    evaluate_parameter_types_from_args,
)
from classiq.model_expansions.function_builder import (
    FunctionContext,
    OperationBuilder,
)
from classiq.model_expansions.scope import Evaluated, QuantumSymbol, Scope
from classiq.model_expansions.utils.counted_name_allocator import CountedNameAllocator
from classiq.qmod.builtins.functions import allocate, free

if TYPE_CHECKING:
    from classiq.model_expansions.interpreter import Interpreter


QuantumStatementT = TypeVar("QuantumStatementT", bound=QuantumStatement)


class Emitter(Generic[QuantumStatementT]):
    def __init__(self, interpreter: "Interpreter") -> None:
        self._interpreter = interpreter

        self._scope_guard = self._interpreter._scope_guard
        self._expand_operation = self._interpreter._expand_operation
        self._machine_precision = self._interpreter._model.preferences.machine_precision

    @abstractmethod
    def emit(self, statement: QuantumStatementT, /) -> None:
        pass

    @property
    def _propagated_var_stack(self) -> PropagatedVarStack:
        return self._interpreter._propagated_var_stack

    @property
    def _builder(self) -> OperationBuilder:
        return self._interpreter._builder

    @property
    def _current_scope(self) -> Scope:
        return self._interpreter._current_scope

    @property
    def _expanded_functions(self) -> List[NativeFunctionDefinition]:
        return self._interpreter._expanded_functions

    @property
    def _counted_name_allocator(self) -> CountedNameAllocator:
        return self._interpreter._counted_name_allocator

    @staticmethod
    def _should_wrap(body: Sequence[QuantumStatement]) -> bool:
        # This protects shadowing of captured variables (i.e, bad user code) by wrapping the body in a function
        # I'm sure there are better ways to handle it, but this is the simplest way to do it for now
        return any(isinstance(stmt, VariableDeclarationStatement) for stmt in body)

    def _create_expanded_wrapping_function(
        self, name: str, body: Sequence[QuantumStatement]
    ) -> QuantumFunctionCall:
        wrapping_function = FunctionClosure.create(
            name=name, body=body, scope=Scope(parent=self._current_scope)
        )
        return self._create_quantum_function_call(wrapping_function, list())

    def _emit_quantum_function_call(
        self, function: FunctionClosure, args: List[ArgValue]
    ) -> QuantumFunctionCall:
        call = self._create_quantum_function_call(function, args)
        self._builder.emit_statement(call)
        return call

    def _create_quantum_function_call(
        self, function: FunctionClosure, args: List[ArgValue]
    ) -> QuantumFunctionCall:
        evaluated_args = [self._interpreter.evaluate(arg) for arg in args]
        new_declaration = self._prepare_fully_typed_declaration(
            function, evaluated_args
        )
        new_positional_arg_decls = new_declaration.positional_arg_declarations
        is_atomic = function.is_atomic
        if not is_atomic:  # perform monomorphization
            self._add_params_to_scope(
                new_positional_arg_decls, evaluated_args, function
            )
            context = self._expand_operation(
                function.with_new_declaration(new_declaration)
            )
            self._expanded_functions.append(
                self._builder.create_definition(cast(FunctionContext, context))
            )
        new_positional_args = self._get_new_positional_args(
            evaluated_args, is_atomic, new_positional_arg_decls
        )
        new_call = QuantumFunctionCall(
            function=(
                new_declaration.name if is_atomic else self._expanded_functions[-1].name
            ),
            positional_args=new_positional_args,
        )
        is_allocate_or_free = (
            new_call.func_name == allocate.func_decl.name
            or new_call.func_name == free.func_decl.name
        )
        parameters = {
            arg_decl.name: FunctionDebugInfo.param_controller(value=valuated_arg.value)
            for arg_decl, valuated_arg in zip(new_positional_arg_decls, evaluated_args)
            if isinstance(arg_decl, ClassicalParameterDeclaration)
        }

        self._interpreter._model.debug_info[new_call.uuid] = FunctionDebugInfo(
            name=new_call.func_name,
            level=OperationLevel.QMOD_FUNCTION_CALL,
            parameters=parameters,
            is_allocate_or_free=is_allocate_or_free,
        )
        if is_atomic:
            new_call.set_func_decl(new_declaration)
        return new_call

    @staticmethod
    def _add_params_to_scope(
        parameters: Sequence[PositionalArg],
        arguments: Sequence[Evaluated],
        closure: FunctionClosure,
    ) -> None:
        for parameter, argument in zip(parameters, arguments):
            if isinstance(argument.value, QuantumSymbol):
                assert isinstance(parameter, PortDeclaration)
                closure.scope[parameter.name] = Evaluated(
                    QuantumSymbol(
                        handle=HandleBinding(name=parameter.name),
                        quantum_type=parameter.quantum_type,
                    ),
                    defining_function=closure,
                )
            else:
                closure.scope[parameter.name] = argument

    def _get_new_positional_args(
        self,
        evaluated_args: List[Evaluated],
        is_atomic: bool,
        new_positional_arg_decls: Sequence[PositionalArg],
    ) -> List[ArgValue]:
        evaluated_args = add_information_from_output_arguments(
            new_positional_arg_decls, evaluated_args
        )
        if is_atomic:
            return [arg.emit() for arg in evaluated_args]

        positional_args = [
            arg.emit() for arg in evaluated_args if isinstance(arg.value, QuantumSymbol)
        ]

        propagated_variables = self._propagated_var_stack.get_propagated_variables()
        validate_args_are_not_propagated(positional_args, propagated_variables)
        positional_args.extend(propagated_variables)

        return positional_args

    def _prepare_fully_typed_declaration(
        self, function: FunctionClosure, evaluated_args: List[Evaluated]
    ) -> NamedParamsQuantumFunctionDeclaration:
        """
        Given, for example,
        def my_func(x: int, q: QArray["x"], p: QArray[]) -> None:
        ...
        def main(...):
            ...
            allocate(5, s)
            my_func(3, r, s)
        The code below will evaluate x to be 3, q to be of size 3 and p to be of size 5.
        Note that it requires a scope for the parameter declaration space, which is
        different from the call scope. For example, the former uses r,s and the latter
        uses p, q.
        """
        with self._scope_guard(Scope(parent=self._current_scope)):
            # The signature scope is passed as a separate argument to avoid contaminating the statement execution scope
            return NamedParamsQuantumFunctionDeclaration(
                name=function.name,
                positional_arg_declarations=evaluate_parameter_types_from_args(
                    function,
                    function.signature_scope,
                    evaluated_args,
                ),
            )
