from typing import Union

import sympy

from classiq.interface.exceptions import ClassiqExpansionError
from classiq.interface.generator.expressions.evaluated_expression import (
    EvaluatedExpression,
)
from classiq.interface.generator.expressions.expression import Expression
from classiq.interface.generator.functions.builtins.internal_operators import (
    POWER_OPERATOR_NAME,
)
from classiq.interface.model.power import Power

from classiq.model_expansions.closure import Closure
from classiq.model_expansions.quantum_operations.emitter import Emitter
from classiq.model_expansions.scope import Scope


class PowerEmitter(Emitter[Power]):
    _power_value: Union[int, sympy.Basic]
    _power_expr: Expression

    def emit(self, power: Power, /) -> None:
        self._power = power
        self._power_value = self._get_power_value()
        self._power_expr = Expression(
            expr=str(self._power_value), source_ref=power.power.source_ref
        )
        self._power_expr._evaluated_expr = EvaluatedExpression(value=self._power_value)

        if len(power.body) > 1 and isinstance(self._power_value, sympy.Basic):
            self._emit_wrapped()
            return

        self._emit_as_operation(power)

    def _emit_as_operation(self, power: Power) -> None:
        power_operation = Closure(
            name=POWER_OPERATOR_NAME,
            blocks=dict(body=self._power.body),
            scope=Scope(parent=self._current_scope),
        )
        with self._propagated_var_stack.capture_variables(self._power):
            context = self._expand_operation(power_operation)
        self._builder.emit_statement(
            Power(
                body=context.statements("body"),
                power=self._power_expr,
                source_ref=power.source_ref,
            )
        )

    def _emit_wrapped(self) -> None:
        with self._propagated_var_stack.capture_variables(self._power):
            wrapping_function = self._create_expanded_wrapping_function(
                POWER_OPERATOR_NAME, self._power.body
            )
        self._builder.emit_statement(
            Power(
                body=[wrapping_function],
                power=self._power_expr,
                source_ref=self._power.source_ref,
            )
        )

    def _get_power_value(self) -> Union[int, sympy.Basic]:
        power_value = self._interpreter.evaluate(self._power.power).value
        if not (isinstance(power_value, int) or power_value.is_symbol):
            raise ClassiqExpansionError(
                f"`power`'s argument should be an integer or identifier. Complex "
                f"expressions are not supported. Got {str(power_value)!r}"
            )
        return power_value
