import ast
from typing import Tuple

from classiq.interface.exceptions import ClassiqExpansionError
from classiq.interface.generator.compiler_keywords import INPLACE_ARITH_AUX_VAR_PREFIX
from classiq.interface.generator.expressions.expression import Expression
from classiq.interface.model.handle_binding import HandleBinding
from classiq.interface.model.inplace_binary_operation import (
    BinaryOperation,
    InplaceBinaryOperation,
)
from classiq.interface.model.quantum_expressions.arithmetic_operation import (
    ArithmeticOperation,
)
from classiq.interface.model.quantum_expressions.quantum_expression import (
    QuantumAssignmentOperation,
)
from classiq.interface.model.quantum_function_call import QuantumFunctionCall
from classiq.interface.model.quantum_type import (
    QuantumNumeric,
)
from classiq.interface.model.variable_declaration_statement import (
    VariableDeclarationStatement,
)
from classiq.interface.model.within_apply_operation import WithinApply

from classiq.model_expansions.evaluators.quantum_type_utils import copy_type_information
from classiq.model_expansions.quantum_operations.expression_operation import (
    ExpressionOperationEmitter,
)
from classiq.model_expansions.scope import QuantumSymbol
from classiq.model_expansions.visitors.boolean_expression_transformers import (
    BooleanExpressionFuncLibAdapter,
    BooleanExpressionOptimizer,
)
from classiq.qmod import builtins
from classiq.qmod.builtins.functions import X


class QuantumAssignmentOperationEmitter(
    ExpressionOperationEmitter[QuantumAssignmentOperation]
):
    def emit(self, op: QuantumAssignmentOperation, /) -> None:
        new_expression = self._evaluate_op_expression(op)
        arrays_with_subscript = self._get_symbols_to_split(new_expression)
        if len(arrays_with_subscript) > 0:
            self._emit_with_split(op, new_expression, arrays_with_subscript)
            return

        self._emit_op(new_expression, op)

    def _emit_op(self, expression: Expression, op: QuantumAssignmentOperation) -> None:
        op = self._evaluate_types_in_expression(op, expression)
        with self._propagated_var_stack.capture_variables(op):
            pass  # This propagates the expression variables
        if isinstance(op, ArithmeticOperation):
            self._emit_arithmetic_op(op, expression)
            return
        self._emit_general_assignment_operation(op)

    def _emit_arithmetic_op(
        self, op: ArithmeticOperation, expression: Expression
    ) -> None:
        op, expression, is_bool_opt = self._optimize_boolean_expression(op, expression)
        if op.inplace_result:
            self._emit_inplace_arithmetic_op(op, expression, is_bool_opt)
        else:
            self._emit_general_assignment_operation(op)

    def _emit_inplace_arithmetic_op(
        self, op: ArithmeticOperation, expression: Expression, is_bool_opt: bool
    ) -> None:
        if op.result_type.size_in_bits > 1 or not _is_res_boolean(op):
            _validate_naive_inplace_handles(op)
            self._build_naive_inplace(op, expression)
            return

        op, expression, to_invert = self._adapt_boolean_inplace(
            op, expression, is_bool_opt
        )
        self._emit_general_assignment_operation(op)
        if not to_invert:
            return

        call = QuantumFunctionCall(
            function=builtins.functions.X.__name__,  # type:ignore[attr-defined]
            positional_args=[op.result_var],
            source_ref=op.source_ref,
        )
        call.set_func_decl(X.func_decl)
        self._interpreter.emit_statement(call)

    def _emit_general_assignment_operation(
        self, op: QuantumAssignmentOperation
    ) -> None:
        result = self._interpreter.evaluate(op.result_var).as_type(QuantumSymbol)
        copy_type_information(op.result_type, result.quantum_type, str(result.handle))
        self._builder.emit_statement(op)

    def _optimize_boolean_expression(
        self, op: ArithmeticOperation, expression: Expression
    ) -> Tuple[ArithmeticOperation, Expression, bool]:
        if not _all_vars_boolean(op):
            return op, expression, False
        optimizer = BooleanExpressionOptimizer()
        optimized_expression = Expression(
            expr=ast.unparse(optimizer.visit(ast.parse(expression.expr)))
        )
        optimized_expression = self._evaluate_expression(optimized_expression)
        optimized_op = op.copy(update=dict(expression=optimized_expression))
        return optimized_op, optimized_expression, optimizer.is_convertible

    def _adapt_boolean_inplace(
        self, op: ArithmeticOperation, expression: Expression, is_bool_opt: bool
    ) -> Tuple[ArithmeticOperation, Expression, bool]:
        adapter = BooleanExpressionFuncLibAdapter(is_bool_opt)
        adapted_expression = self._evaluate_expression(
            Expression(expr=ast.unparse(adapter.visit(ast.parse(expression.expr))))
        )
        adapted_op = op.copy(update=dict(expression=adapted_expression))
        return adapted_op, adapted_expression, adapter.to_invert

    def _build_naive_inplace(
        self, qe: ArithmeticOperation, new_expression: Expression
    ) -> None:
        aux_var = self._counted_name_allocator.allocate(INPLACE_ARITH_AUX_VAR_PREFIX)
        self._interpreter.emit_statement(
            VariableDeclarationStatement(name=aux_var, quantum_type=QuantumNumeric())
        )
        arith_expression = ArithmeticOperation(
            result_var=HandleBinding(name=aux_var),
            expression=new_expression,
            inplace_result=False,
        )
        inplace_store = InplaceBinaryOperation(
            operation=BinaryOperation.Xor,
            target=qe.result_var,
            value=HandleBinding(name=aux_var),
        )
        self._interpreter.emit_statement(
            WithinApply(compute=[arith_expression], action=[inplace_store])
        )


def _validate_naive_inplace_handles(qe: ArithmeticOperation) -> None:
    if qe.result_var in qe.var_handles:
        raise ClassiqExpansionError(
            f"Quantum variable {qe.result_var.name} cannot appear on both sides of "
            f"the ^= assignment"
        )


def _all_vars_boolean(op: ArithmeticOperation) -> bool:
    if not all(
        var_type.has_size_in_bits and var_type.size_in_bits == 1
        for var_type in op.var_types.values()
    ):
        return False
    if any(
        isinstance(var_type, QuantumNumeric)
        and (var_type.sign_value or var_type.fraction_digits_value > 0)
        for var_type in op.var_types.values()
    ):
        return False
    return _is_res_boolean(op)


def _is_res_boolean(op: ArithmeticOperation) -> bool:
    if not (op.result_type.has_size_in_bits and op.result_type.size_in_bits == 1):
        return False
    return not (
        isinstance(op.result_type, QuantumNumeric)
        and (op.result_type.sign_value or op.result_type.fraction_digits_value > 0)
    )
